import numpy as np
import time
import os

from .. import takagitaupin

def bragg_reflectivity_sigma():
    print ('Computing Bragg reflectivity for 300 um thick, non-bent GaAS(400)')
    print ('for sigma polarized beam at 6 keV in angle domain.')

    th = np.linspace(-10,30,150)

    t0 = time.time()
    R,T=takagitaupin('angle',th,6,'sigma','GaAs',[4,0,0],0,300,None,1,1e-10)
    print('It took ' + str(time.time()-t0) + ' seconds to solve ' + \
          str(th.size) + ' points.\n')


    A=np.loadtxt(os.path.dirname(__file__)+'/reference_curves/GaAs400_6keV_300micron_unbent_sigma.dat')

    th_ref = A[:,0]
    R_ref = A[:,1]


    return th, R, th_ref, R_ref

def bragg_reflectivity_pi_asymmetric():
    print ('Computing Bragg reflectivity for 20 um thick, non-bent Li(220) with')
    print ('asymmetry of 10 deg for pi polarized beam at 8 keV in angle domain.')

    th = np.linspace(-2,8,150)

    t0 = time.time()
    R,T=takagitaupin('angle',th,8,'pi','LiF',[2,2,0],10,20,None,1,1e-10)
    print('It took ' + str(time.time()-t0) + ' seconds to solve ' + \
          str(th.size) + ' points.\n')

    A=np.loadtxt(os.path.dirname(__file__)+'/reference_curves/LiF220_8keV_20micron_unbent_pi_asymm_10.dat')

    th_ref = A[:,0]
    R_ref = A[:,1]

    return th, R, th_ref, R_ref

def laue_symmetric():
    print ('Computing Laue reflectivity and transmission for 100 um thick')
    print ('non-bent Ge(111) for sigma polarized beam at 6 keV in angle domain.')

    th = np.linspace(-40,20,150)

    t0 = time.time()
    R,T=takagitaupin('angle',th,6,'sigma','Ge',[1,1,1],90,100,None,1,1e-11)
    print('It took ' + str(time.time()-t0) + ' seconds to solve ' + \
          str(th.size) + ' points.\n')

    A=np.loadtxt(os.path.dirname(__file__)+'/reference_curves/Ge111_6keV_100micron_unbent_laue_refl.dat')
    B=np.loadtxt(os.path.dirname(__file__)+'/reference_curves/Ge111_6keV_100micron_unbent_laue_trans.dat')

    th_ref = A[:,0]
    R_ref = A[:,1]
    T_ref = B[:,1]

    return th, R, T, th_ref, R_ref, T_ref


def bragg_bent_symmetric():
    print ('Computing Bragg reflectivity for 300 um thick, cylindrically bent')
    print ('Si(660) (1m bending radius) for sigma polarized at 88 deg in energy domain.')

    e = np.linspace(-450,250,150)

    thickness=300
    R_bend = 1
    nu = 0.27

    #Define Jacobian for cylindrical bending
    def ujac2(x,y,thickness,R_bend,nu):
        ux_x = -1/R_bend*(y+thickness/2)
        ux_y = -x/R_bend

        uy_x = x/R_bend
        uy_y = nu/R_bend*(thickness/2+y)

        return np.array([[ux_x,ux_y],[uy_x,uy_y]])

    def ujac(x,y):
        return ujac2(x,y,thickness*1e-6,R_bend,nu)


    t0 = time.time()
    R,T=takagitaupin('energy',e,88,'sigma','Si',[6,6,0],0,thickness,ujac,1,1e-10)
    print('It took ' + str(time.time()-t0) + ' seconds to solve ' + \
          str(e.size) + ' points.\n')

    #TODO: ADD REFERENCE FOR BENT (multilamellar?)

    return e, R




from .examples import *
import matplotlib.pyplot as plt
import time

__all__ = ['run']

def run():
    print('Running all examples.')
    t0 = time.time()
    res = bragg_reflectivity_sigma()
    res2 = bragg_reflectivity_pi_asymmetric()
    res3 = laue_symmetric()
    res4 = bragg_bent_symmetric()
    print('Total running time: ' + str(time.time()-t0) + ' seconds.')

    plt.figure()
    plt.plot(res[0],res[1],linewidth = 2,label='pyTT')
    plt.plot(res[2],res[3],linewidth = 2,label='XINPRO 1.2\n(T=25 K)')

    plt.title('Unbent GaAs (004) in Bragg geometry')
    plt.xlabel('$\\theta-\\theta_B$ (arc sec)')
    plt.ylabel('Reflectivity')
    plt.legend()

    plt.figure()
    plt.plot(res2[0],res2[1],linewidth = 2,label='pyTT')
    plt.plot(res2[2],res2[3],linewidth = 2,label='XINPRO 1.2\n(T=75 K)')

    plt.title('Unbent LiF (220) with 10 deg asymmetry in Bragg geometry')
    plt.xlabel('$\\theta-\\theta_B$ (arc sec)')
    plt.ylabel('Reflectivity')
    plt.legend()

    plt.figure()
    plt.plot(res3[0],res3[1],linewidth = 2,label='pyTT (reflected)')
    plt.plot(res3[0],res3[2],linewidth = 2,label='pyTT (transmitted)')
    plt.plot(res3[3],res3[4],linewidth = 2,label='XINPRO 1.2\n(reflected, T=50 K)')
    plt.plot(res3[3],res3[5],linewidth = 2,label='XINPRO 1.2\n(transmitted, T=50 K)')

    plt.title('Unbent Ge (111) in Laue geometry')
    plt.xlabel('$\\theta-\\theta_B$ (arc sec)')
    plt.ylabel('Reflectivity')
    plt.legend()

    plt.figure()
    plt.plot(res4[0],res4[1],linewidth = 2,label='pyTT')

    plt.title('Cylindrically bent (R_b = 1 m) Si (660) in Bragg geometry at 88 deg')
    plt.xlabel('$\Delta E$ (meV)')
    plt.ylabel('Reflectivity')
    plt.legend()


    plt.show()



from __future__ import division, print_function
import sys
import os

import numpy as np
import matplotlib.pyplot as plt

from scipy.integrate import ode
from scipy.constants.codata import physical_constants

import xraylib

def takagitaupin(scantype,scan,constant,polarization,crystal_str,hkl,asymmetry,thickness,displacement_jacobian = None,debyeWaller=1.0,min_int_step=1e-10):
    '''
    1D TT-solver.
    
    Input:
    scantype = 'energy' or 'angle'
    scan =  relative to the Bragg's energy in meV (energy scan) OR relative to the Bragg's angle in arcsec (angle scan). 
            scan is an numpy array containing these values OR an integer N generating automatic scan with N points. 
    constant = incidence angle respect to the diffracting planes in degrees (energy scan) OR photon energy in keV (angle scan) 
    polarization = 'sigma' or 'pi'
    crystal_str = supports crystals included in xraylib, e.g. 'Si', 'Ge', 'LiF'
    hkl = [h,k,l] (Miller indices)
    asymmetry = asymmetry angle (in deg, positive to clockwise direction)
    thickness = crystal thickness in microns
    displacement_jacobian = a function giving the jacobian of the displacement field as function of position (x,y). 
                            Note: y points upwards the crystal 
    debyeWaller = Debye-Waller factor
    min_int_step = minumum integration step
    '''

    if scantype == 'energy':
        is_escan = True
        scantype = 'energy'
    elif scantype == 'angle':
        is_escan = False
        scantype = 'angle'


    if scantype == 'energy':
        print('Computing elastic line for ' + str(hkl) + ' reflection of ' \
              + crystal_str + ' crystal in energy domain.' )
        is_escan = True
    elif scantype == 'angle':
        print('Computing elastic line for ' + str(hkl) + ' reflection of ' \
              + crystal_str + ' crystal in angle domain.' )
        is_escan = False

    #type conversions
    if type(scan) is int:
        print('AUTOMATIC LIMITS NOT IMPLEMENTED YET!')
        print('Function terminated.')
        return None

    scan=np.array(scan)
    
    #Unit conversions
    thickness = thickness*1e-6 #wafer thickness in meters

    #constants
    crystal = xraylib.Crystal_GetCrystal(crystal_str)

    hc = physical_constants['Planck constant in eV s'][0]*physical_constants['speed of light in vacuum'][0]*1e3 #in meV*m
    d = xraylib.Crystal_dSpacing(crystal,*hkl)*1e-10 #in m
    V = crystal['volume']*1e-30 # volume of unit cell in m^3
    r_e = physical_constants['classical electron radius'][0]

    print('')
    print('Crystal     : ', crystal_str)
    print('Reflection  : ', hkl)
    print('d_hkl       : ', d, ' m')
    print('Cell volume : ', V, ' m^3')
    print('')


    #asymmetry angle
    phi=np.radians(asymmetry)

    #Setup scan variables and constants
    if is_escan:
        escan=scan

        th0=np.radians(constant)
        th=th0

        #Conversion of incident photon energy to wavelength
        E0 = hc/(2*d*np.sin(th)) #in meV

        wavelength = hc/(E0+escan) #in m
        k = 2*np.pi/wavelength #in 1/m

    else:
        E0 = constant*1e6 #in meV

        wavelength = hc/E0 #in m
        k = 2*np.pi/wavelength #in 1/m

        if not hc/(2*d*E0) > 1:
            th0 = np.arcsin(hc/(2*d*E0))
        else:
            print('Given energy below the backscattering energy!')
            print('Setting theta to 90 deg.')
            th0 = np.pi/2

        ascan = scan*np.pi/648000 #from arcsec to rad
        th = th0+ascan

    #Incidence and exit angles
    alpha0 = th+phi
    alphah = th-phi
    
    #Direction parameters
    gamma0 = np.ones(scan.shape)/np.sin(alpha0)
    gammah = np.ones(scan.shape)/np.sin(alphah)

    if np.mean(gammah) < 0:
        print('The direction of diffraction in to the crystal -> Laue case')
        geometry = 'laue'
    else:
        print('The direction of diffraction out of the crystal -> Bragg case')
        geometry = 'bragg'

    #Polarization
    if polarization == 'sigma':
        C = 1;
        print('Solving for sigma-polarization')
    else:
        C = np.cos(2*th);
        print('Solving for pi-polarization')

    print('Asymmetry angle : ', phi,' rad, ', np.degrees(phi), ' deg')
    print('Wavelength      : ', hc/E0*1e10, ' Angstrom ')
    print('Energy          : ', E0*1e-6, ' keV ')
    
    print('Bragg angle     : ', th0,' rad, ', np.degrees(th0), ' deg')
    print('Incidence angle : ', th0+phi,' rad ', np.degrees(th0+phi), ' deg')
    print('Exit angle      : ', th0-phi,' rad ', np.degrees(th0-phi), ' deg')
    print('')

    #Compute susceptibilities
    if is_escan:
        F0 = np.zeros(escan.shape,dtype=np.complex)
        Fh = np.zeros(escan.shape,dtype=np.complex)
        Fb = np.zeros(escan.shape,dtype=np.complex)

        for ii in range(escan.size):    
            F0[ii] = xraylib.Crystal_F_H_StructureFactor(crystal, (E0+escan[ii])*1e-6, 0, 0, 0, debyeWaller, 1.0)
            Fh[ii] = xraylib.Crystal_F_H_StructureFactor(crystal, (E0+escan[ii])*1e-6, hkl[0], hkl[1], hkl[2], debyeWaller, 1.0)
            Fb[ii] = xraylib.Crystal_F_H_StructureFactor(crystal, (E0+escan[ii])*1e-6, -hkl[0], -hkl[1], -hkl[2], debyeWaller, 1.0)
    else:
        F0 = xraylib.Crystal_F_H_StructureFactor(crystal, E0*1e-6, 0, 0, 0, debyeWaller, 1.0)
        Fh = xraylib.Crystal_F_H_StructureFactor(crystal, E0*1e-6, hkl[0], hkl[1], hkl[2], debyeWaller, 1.0)
        Fb = xraylib.Crystal_F_H_StructureFactor(crystal, E0*1e-6, -hkl[0], -hkl[1], -hkl[2], debyeWaller, 1.0)

    cte = - r_e * wavelength*wavelength/(np.pi * V)
    chi0 = np.conj(cte*F0)
    chih = np.conj(cte*Fh)
    chib = np.conj(cte*Fb)

    print('F0   : ',np.mean(F0))
    print('Fh   : ',np.mean(Fh))
    print('Fb   : ',np.mean(Fb))
    print('')
    print('chi0 : ',np.mean(chi0))
    print('chih : ',np.mean(chih))
    print('chib : ',np.mean(chib))
    print('')
    print('(Mean F and chi values for energy scan)')
    print('')

    ######################
    #COEFFICIENTS FOR TTE#
    ######################

    #For solving ksi = Dh/D0 
    c0 = k*chi0/2*(gamma0+gammah)*np.ones(scan.shape)
    ch = k*C*chih*gammah/2*np.ones(scan.shape)
    cb = k*C*chib*gamma0/2*np.ones(scan.shape)

    #For solving Y = D0 
    g0 = k*chi0/2*gamma0*np.ones(scan.shape)
    gb = k*C*chib/2*gamma0*np.ones(scan.shape)

    #deviation from the kinematical Bragg condition for unstrained crystal
    beta = 2*np.pi/d*gammah*(np.sin(th)-wavelength/(2*d))

    #Bending
    if not displacement_jacobian == None:

        def strain_term(z,step):
            if is_escan:
                x = -z*np.cos(alpha0)/np.sin(alpha0)
                u_jac = displacement_jacobian(x,z)
                duh_dsh = 2*np.pi/d*(np.sin(phi)*np.cos(alphah)*u_jac[0,0] 
                                    +np.sin(phi)*np.sin(alphah)*u_jac[0,1]
                                    +np.cos(phi)*np.cos(alphah)*u_jac[1,0]
                                    +np.cos(phi)*np.sin(alphah)*u_jac[1,1]
                                    )
                return gammah[step]*duh_dsh
            else:
                x = -z*np.cos(alpha0[step])/np.sin(alpha0[step])
                u_jac = displacement_jacobian(x,z)
                duh_dsh = 2*np.pi/d*(np.sin(phi)*np.cos(alphah[step])*u_jac[0,0]
                                    +np.sin(phi)*np.sin(alphah[step])*u_jac[0,1]
                                    +np.cos(phi)*np.cos(alphah[step])*u_jac[1,0] 
                                    +np.cos(phi)*np.sin(alphah[step])*u_jac[1,1]
                                    )
                return gammah[step]*duh_dsh

    else:
        def strain_term(z,step): 
            return 0

    #INTEGRATION

    #Define ODEs and their Jacobians
    if geometry == 'bragg':
        print('Transmission in the Bragg case not implemented!')

        reflectivity = np.zeros(scan.shape)
        transmission = -np.ones(scan.shape)

        def ksiprime(z,ksi,step):
            return 1j*cb[step]*ksi**2+1j*(c0[step]+beta[step]+strain_term(z,step))*ksi+1j*ch[step]

        def ksiprime_jac(z,ksi,step):
            return 2j*cb[step]*ksi+1j*(c0[step]+beta[step]+strain_term(z,step))

    else:

        forward_diffraction = np.zeros(scan.shape)
        diffraction = np.zeros(scan.shape)

        def TTE(z,Y,step):
            return [1j*cb[step]*Y[0]**2+1j*(c0[step]+beta[step]+strain_term(z,step))*Y[0]+1j*ch[step],\
                    -1j*(g0[step]+gb[step]*Y[0])*Y[1]]

        def TTE_jac(z,Y,step):
            return [[2j*cb[step]*Y[0]+1j*(c0[step]+beta[step]+strain_term(z,step)), 0],\
                    [-1j*gb[step]*Y[1],-1j*(g0[step]+gb[step]*Y[0])]]


    #Solve the equation
    sys.stdout.write('Solving...0%')
    sys.stdout.flush()
    
    for step in range(len(scan)):
        if geometry == 'bragg':
            r=ode(ksiprime,ksiprime_jac)
        else:
            r=ode(TTE,TTE_jac)

        r.set_integrator('zvode',method='bdf',with_jacobian=True, min_step=min_int_step,max_step=1e-4,nsteps=50000)
        r.set_f_params(step)
        r.set_jac_params(step)
        
        if geometry == 'bragg':
            r.set_initial_value(0,-thickness)
            res=r.integrate(0)     
            reflectivity[step]=np.abs(res[0])**2*gamma0[step]/gammah[step] #gamma-part takes into account beam footprints 
        else:
            r.set_initial_value([0,1],0)
            res=r.integrate(-thickness)
            diffraction[step] = np.abs(res[0]*res[1])**2
            forward_diffraction[step] = np.abs(res[1])**2

        sys.stdout.write('\rSolving...%0.1f%%' % (100*(step+1)/len(scan),))  
        sys.stdout.flush()

    sys.stdout.write('\r\nDone.\n')
    sys.stdout.flush()

    if geometry == 'bragg':    
        return reflectivity, transmission
    else:    
        return diffraction, forward_diffraction




__author__ = 'Ari-Pekka Honkanen'
__license__ = 'MIT'
__date__ = '17/11/2017'

from setuptools import setup
#
# memorandum (for pypi)
#
# python setup.py sdist upload


setup(name='pyTTE',
      version='0.1.3',
      description='Package to compute X-ray diffraction curves of bent crystals by numerically solving the Takagi-Taupin equations.',
      author='Ari-Pekka Honkanen',
      author_email='honkanen.ap@gmail.com',
      url='https://github.com/aripekka/pyTTE/',
      packages=[
                'pyTTE',
                'pyTTE.examples',
               ],
      install_requires=[
                        'numpy',
                        'scipy',
                        'matplotlib'
                       ],
      include_package_data=True,
)
